/* Embedded OS Loader
 *
 *   Copyright 2005 Vivien Chappelier <vivien.chappelier@free.fr>
 *
 * miniSD I/O
 *
 */
#include "stdlib.h"
#include "mmc.h"

/* initialize data transfer parameters */
void mmc_data_init(int sector_size, int sector_count)
{
  MMC_WRITE(BUF, 0x1f00); /* buffer full at 31, empty at 0, dma disabled */
  MMC_WRITE(BLEN, sector_size - 1);
  MMC_WRITE(NBLK, sector_count - 1);
}

/* send a command to the mmc controller */
int mmc_send_command(int command, int arg)
{
  MMC_WRITE(CTO, 200); /* timeout of 200 cycles */

  /* send the argument */
  MMC_WRITE(ARGL, arg & 0xffff);
  MMC_WRITE(ARGH, arg >> 16);

  /* send the command */
  MMC_WRITE(CMD, command);
  while(!(MMC_READ(STAT) & (MMC_STAT_CMD_TOUT | MMC_STAT_END_OF_CMD | MMC_STAT_CARD_ERR)));
  udelay(10);
  if(MMC_READ(STAT) & MMC_STAT_CMD_TOUT) {
    /* timeout */
    MMC_WRITE(STAT, MMC_STAT_CMD_TOUT); /* ack */
    printb("mmc_send_command: timeout\n");
    return(0); 
  }
  if(MMC_READ(STAT) & MMC_STAT_CARD_ERR) {
    /* card error: read response register */
    int response = MMC_READ(RSP6) | (MMC_READ(RSP7) << 16);
    MMC_WRITE(STAT, MMC_STAT_CARD_ERR); /* ack */
    printb("mmc_send_command: card error %x\n", response);
    printb("mmc: ARGL = %x\n", MMC_READ(ARGL));
    printb("mmc: ARGH = %x\n", MMC_READ(ARGH));
    return(0);
  }

  MMC_WRITE(STAT, MMC_STAT_END_OF_CMD); /* ack */
  //TEMP
  //  printb("mmc_send_command: success\n");
  return(1);
}

/* read some data from mmc controller buffer in PIO mode */
/* count is in sectors of 512 bytes */
int mmc_read_data(void *buffer, int count)
{
  int i, j, n;
  unsigned short *data = (unsigned short *) buffer;

  if(count > 0) {
    n = count * SECTOR_SIZE / 64; /* read by chunks of 64 bytes */

    for(j = 0; j < n; j++) {
      // TEMP
      //      printb(".");
      while(!(MMC_READ(STAT) & MMC_STAT_A_FULL)); /* wait for data */
      MMC_WRITE(STAT, MMC_STAT_A_FULL); /* ack */
      for(i = 0; i < 32; i++) /* 32 short words = 64 bytes */
	*data++ = MMC_READ(DATA);
    }
  }

  //TEMP
  //  printb("\n"); 
  while(!(MMC_READ(STAT) & MMC_STAT_END_OF_DATA)); /* wait for data */
  MMC_WRITE(STAT, MMC_STAT_END_OF_DATA); /* ack */
  //TEMP
  //  printb("transfer done\n"); 

  return(1);
}

/* read 512-byte sectors from the MMC */
int mmc_read_sectors(int sector_offset, void *buffer, size_t sector_count)
{
  if(sector_count <= 0) return(1);

  mmc_data_init(SECTOR_SIZE, sector_count);

  // TEMP
  //  printb("mmc_read_sector %x %x: STAT = %x\n", sector_offset, sector_count, MMC_READ(STAT));

  if(sector_count > 1) { /* multiple block request */
    mmc_send_command(MMC_CMD_READ_MULTIPLE_BLOCK, sector_offset * SECTOR_SIZE);
    mmc_read_data(buffer, sector_count);
    mmc_send_command(MMC_CMD_STOP_TRANSMISSION, 0);
  } else {               /* single block request   */
    mmc_send_command(MMC_CMD_READ_SINGLE_BLOCK, sector_offset * SECTOR_SIZE);
    mmc_read_data(buffer, sector_count);
  }
  
  return(1);
}

/* read data from the MMC */
int mmc_read(int offset, void *buffer, size_t length)
{
  unsigned char *data  = (unsigned char *) buffer;
  unsigned char sector_data[SECTOR_SIZE];
  int start_length = (SECTOR_SIZE - (offset & SECTOR_MASK)) & SECTOR_MASK;
  int end_length = (offset + length) & SECTOR_MASK;
  int start_sector = offset >> SECTOR_SHIFT;
  int end_sector = (offset + length) >> SECTOR_SHIFT;
  int sector_count = (length - start_length) >> SECTOR_SHIFT;

  if(start_length) {
    if(!mmc_read_sectors(start_sector, sector_data, 1))
      return(0);
    memcpy(data, sector_data + (offset & SECTOR_MASK), start_length);
    data += start_length;
    start_sector++;
  }

  if(!mmc_read_sectors(start_sector, data, sector_count))
    return(start_length);
  data += sector_count * SECTOR_SIZE;

  if(end_length) {
    if(!mmc_read_sectors(end_sector, sector_data, 1))
      return(sector_count * SECTOR_SIZE + start_length);
    memcpy(data, sector_data, end_length);
  }

  return(length);
}

/* reset the MMC controller */
int mmc_reset()
{
  /* clear the status register */
  MMC_WRITE(STAT, 0xffff);

  /* send initialisation stream/data abort command */
  mmc_send_command(MMC_CMD_RESET, 0);

  return(MMC_READ(STAT));
}
